; Midi/Karaoke player V4 for 6502+6522+6551
; (C) F.Frances 2004

#define DEBUG 0

;***************************************************************
; Oric-specific definitions

VIA = $0300
ACIA = $031C
xvdu = $0238
text_status = $026A
ink = $026B
paper = $026C
newline	= $CBF0	; CR+LF
clearscreen = $CCCE
irq_vector = $244		; JMP to irq handler
error = $C47E		; error routine
nocaps = $F766		; clear CAPS message

FILE_START=$0600	; must be page-aligned

;***************************************************************
; Oric page 0 allocations

melody	= 0
screen_line = $12		; 2 bytes ptr of screen line

txtbuf	= $35		; safe location for page 0 variables
txtbuflen	= 80

delta		= $38		; temporary variable (lsb)
ysave		= $39		; to save Y register
tempo		= $3A		; 3 bytes value
xsave		= $3F		; to save X register
track_ptr	= $40-2	; 24*2 contiguous bytes for the tracks pointers
fileptr	= $70		; 2 bytes ptr
tmp		= $72		; 2 bytes value
length	= $72		; 2 bytes value
nbtracks2	= $74		; twice the number of tracks
status_send	= $75		; flag indicating if status must be sent
ppqn		= $76		; 2 bytes Pulse Per Quarter Note
no_active_tracks = $78	; counter indicating how many tracks are active
clock_pulse	= $79		; flag indicating if next clock pulse has occured
irq_save	= $7A		; address of old irq handler
current_channel = $7C	; status of the current message
ptr_first	= $7D		; offset in the transmit buffer
ptr_last	= $7E		; offset in the transmit buffer
cursor	= $7F		; offset for text display
status_save	= $80		; to save ACIA_STATUS
counter	= $81		; soft counter for clock pulses
counter_start = $82	; initial value of clock pulse counter
checked	= $83		; flag indicating if next text message has been tested

;***************************************************************
; use page 1 for arrays
delta_counter	= $0110-2	; 24 words for the delta-time counters
track_status	= $0140-2	; 24 bytes for the individual tracks status
delta_high		= $0141-2	; 24 bytes for the msb of delta-time counters

;***************************************************************
; use $B400-$B4FF for ACIA transmit buffer,
; and transient initialisation routine

*=$B400
buffer

init
	lda #7
	sta ink
	lda #$10
	sta paper
	lda #$22		; 40 columns, no attributes
	sta text_status
	ldx #$0C
	jsr xvdu
	jsr nocaps
	jsr clear_vars
	sta fileptr
	tay

	ldx #3
init_vars
	lda default_tempo-1,x
	sta tempo-1,x
	dex
	bne init_vars
 
	lda #>FILE_START
	sta fileptr+1
check_file_sig
	lda (fileptr),y
	cmp file_signature,y
	bne bad_format
	iny
	cpy #9
	bne check_file_sig
	lda (fileptr),y
	cmp #2
	bcs bad_format	; midi format types 0 and 1 accepted
	iny
	lda (fileptr),y
	bne bad_format
	iny
	lda (fileptr),y
	iny
	cmp #25
	bcs bad_format	; 24 tracks max
	asl
	sta nbtracks2
	lda (fileptr),y
	bmi bad_format	; SMPTE format not accepted
	sta ppqn+1
	iny
	lda (fileptr),y
	sta ppqn
	ldx #0
	lda fileptr+1
init_next_track_ptr
	inx
	inx			; tracks numbered from 1 to 24
	sty track_ptr,x
	sta track_ptr+1,x
	lda #$80
	sta track_status,x
	ldy #5		; 6 bytes track signature to check
check_track_sig
	jsr inc_ptr
	cmp track_signature,y
	bne bad_format
	dey
	bpl check_track_sig
	jsr inc_ptr
	sta length+1
	jsr inc_ptr
	sta length
	clc
	lda track_ptr,x
	adc length
	tay
	lda track_ptr+1,x
	adc length+1
	cpx nbtracks2
	bne init_next_track_ptr
	jmp init_hardware

bad_format
	ldx #$31	; DATA ERROR message
	jmp error

init_hardware
	lda #0
	sta ACIA_COMMAND	; disable ACIA
	lda #$7F		; clear all sources of VIA interrupt
	sta VIA_IER
	lda #$C0		; set T1 as the only source of interrupt
	sta VIA_IER
	lda irq_vector+1	; save old irq vector
	sta irq_save
	lda irq_vector+2
	sta irq_save+1
	sei			; set new irq vector
	lda #<irq_handler
	sta irq_vector+1
	lda #>irq_handler
	sta irq_vector+2
	cli
	lda #$10		; set baud rate generator to clock/16 = 31250 baud
	sta ACIA_CONTROL
	lda #$07		; set DTR, RTS, and accept transmit interrupts
	sta ACIA_COMMAND

	jmp init_deltas

track_signature	; backwards
		.db 0
		.db 0
		.db "krTM"
file_signature
		.db "MThd"
		.db 0,0,0,6,0

default_tempo
		.db $20,$A1,$07	; 500000 microseconds = 120 BPM

	.dsb $B500-*,0
;***************************************************************
; $B500-$B7FF charset

 .db $00,$00,$00,$00,$00,$00,$00,$00	;  
 .db $08,$08,$08,$08,$08,$00,$08,$00	; !
 .db $14,$14,$14,$00,$00,$00,$00,$00	; "
 .db $00,$00,$00,$00,$00,$20,$20,$20	; upper part of b,h,k
 .db $00,$00,$00,$00,$00,$02,$02,$02	; upper part of d
 .db $00,$00,$00,$00,$00,$0C,$10,$10	; upper part of f
 .db $1C,$22,$22,$2C,$22,$22,$2C,$00	; free
 .db $08,$08,$08,$00,$00,$00,$00,$00	; '
 .db $08,$10,$20,$20,$20,$10,$08,$00	; (
 .db $08,$04,$02,$02,$02,$04,$08,$00	; )
 .db $00,$00,$00,$00,$22,$14,$08,$00	; accent circonflexe invers
 .db $00,$00,$00,$00,$10,$08,$04,$00	; accent grave
 .db $00,$00,$00,$00,$00,$08,$08,$10	; ,
 .db $00,$00,$00,$3E,$00,$00,$00,$00	; -
 .db $00,$00,$00,$00,$00,$00,$08,$00	; . (and upper part of i,j)
 .db $00,$00,$00,$00,$08,$14,$08,$00	; accent rond

 .db $1C,$22,$26,$2A,$32,$22,$1C,$00	; 0
 .db $08,$18,$08,$08,$08,$08,$1C,$00	; 1
 .db $1C,$22,$02,$04,$08,$10,$3E,$00	; 2
 .db $3E,$02,$04,$0C,$02,$22,$1C,$00	; 3
 .db $04,$0C,$14,$24,$3E,$04,$04,$00	; 4
 .db $3E,$20,$3C,$02,$02,$22,$1C,$00	; 5
 .db $0C,$10,$20,$3C,$22,$22,$1C,$00	; 6
 .db $3E,$02,$04,$08,$10,$10,$10,$00	; 7
 .db $1C,$22,$22,$1C,$22,$22,$1C,$00	; 8
 .db $1C,$22,$22,$1E,$02,$04,$18,$00	; 9
 .db $00,$00,$08,$00,$00,$08,$00,$00	; :
 .db $00,$00,$08,$00,$00,$08,$08,$10	; ;
 .db $00,$00,$00,$00,$00,$00,$08,$00	; free
 .db $00,$00,$00,$00,$00,$18,$08,$08	; upper part of l
 .db $00,$00,$00,$00,$00,$00,$10,$10	; upper part of t
 .db $1C,$22,$04,$08,$08,$00,$08,$00	; ?

 .db $1A,$24,$2A,$2A,$2A,$12,$2C,$00	; 
 .db $08,$14,$22,$22,$3E,$22,$22,$00	; A
 .db $3C,$22,$22,$3C,$22,$22,$3C,$00	; B
 .db $1C,$22,$20,$20,$20,$22,$1C,$00	; C
 .db $3C,$22,$22,$22,$22,$22,$3C,$00	; D
 .db $3E,$20,$20,$3C,$20,$20,$3E,$00	; E
 .db $3E,$20,$20,$3C,$20,$20,$20,$00	; F
 .db $1E,$20,$20,$20,$26,$22,$1E,$00	; G
 .db $22,$22,$22,$3E,$22,$22,$22,$00	; H
 .db $1C,$08,$08,$08,$08,$08,$1C,$00	; I
 .db $02,$02,$02,$02,$02,$22,$1C,$00	; J
 .db $22,$24,$28,$30,$28,$24,$22,$00	; K
 .db $20,$20,$20,$20,$20,$20,$3E,$00	; L
 .db $22,$36,$2A,$2A,$22,$22,$22,$00	; M
 .db $22,$22,$32,$2A,$26,$22,$22,$00	; N
 .db $1C,$22,$22,$22,$22,$22,$1C,$00	; O

 .db $3C,$22,$22,$3C,$20,$20,$20,$00	; P
 .db $1C,$22,$22,$22,$2A,$24,$1A,$00	; Q
 .db $3C,$22,$22,$3C,$28,$24,$22,$00	; R
 .db $1C,$22,$20,$1C,$02,$22,$1C,$00	; S
 .db $3E,$08,$08,$08,$08,$08,$08,$00	; T
 .db $22,$22,$22,$22,$22,$22,$1C,$00	; U
 .db $22,$22,$22,$22,$22,$14,$08,$00	; V
 .db $22,$22,$22,$2A,$2A,$36,$22,$00	; W
 .db $22,$22,$14,$08,$14,$22,$22,$00	; X
 .db $22,$22,$14,$08,$08,$08,$08,$00	; Y
 .db $3E,$02,$04,$08,$10,$20,$3E,$00	; Z
 .db $1E,$28,$28,$3C,$28,$28,$2E,$00	; AE
 .db $1C,$22,$20,$20,$20,$22,$1C,$38	; 
 .db $1C,$22,$22,$2C,$22,$22,$2C,$00	; 
 .db $00,$00,$00,$00,$08,$14,$22,$00	; accent circonflexe
 .db $00,$00,$00,$00,$04,$08,$10,$00	; accent aigu

 .db $1A,$24,$2A,$2A,$12,$2C,$00,$00	; 
 .db $1C,$02,$1E,$22,$22,$1E,$00,$00	; a
 .db $3C,$22,$22,$22,$22,$3C,$00,$00	; b
 .db $1C,$22,$20,$20,$22,$1C,$00,$00	; c
 .db $1E,$22,$22,$22,$22,$1E,$00,$00	; d
 .db $1C,$22,$3E,$20,$22,$1C,$00,$00	; e
 .db $3C,$10,$10,$10,$10,$10,$00,$00	; f
 .db $1C,$22,$22,$22,$22,$1E,$02,$1C	; g
 .db $3C,$22,$22,$22,$22,$22,$00,$00	; h
 .db $18,$08,$08,$08,$08,$1C,$00,$00	; i
 .db $18,$08,$08,$08,$08,$08,$08,$10	; j
 .db $24,$28,$30,$28,$24,$22,$00,$00	; k
 .db $08,$08,$08,$08,$08,$1C,$00,$00	; l
 .db $34,$2A,$2A,$2A,$2A,$22,$00,$00	; m
 .db $3C,$22,$22,$22,$22,$22,$00,$00	; n
 .db $1C,$22,$22,$22,$22,$1C,$00,$00	; o

 .db $3C,$22,$22,$22,$22,$3C,$20,$20	; p
 .db $1E,$22,$22,$22,$22,$1E,$02,$02	; q
 .db $2C,$32,$20,$20,$20,$20,$00,$00	; r
 .db $1C,$22,$18,$04,$22,$1C,$00,$00	; s
 .db $3C,$10,$10,$10,$10,$0C,$00,$00	; t
 .db $22,$22,$22,$22,$26,$1A,$00,$00	; u
 .db $22,$22,$14,$14,$08,$08,$00,$00	; v
 .db $22,$2A,$2A,$2A,$14,$14,$00,$00	; w
 .db $22,$14,$08,$08,$14,$22,$00,$00	; x
 .db $22,$22,$14,$14,$08,$08,$10,$10	; y
 .db $3E,$24,$08,$10,$22,$3E,$00,$00	; z
 .db $34,$0A,$1E,$28,$2A,$14,$00,$00	; ae
 .db $1C,$22,$20,$20,$22,$1C,$08,$30	; 
 .db $1C,$2A,$2E,$28,$2A,$1C,$00,$00	; free
 .db $00,$00,$00,$00,$00,$1A,$16,$00	; tilde
 .db $00,$00,$00,$00,$00,$00,$14,$00	; accent trema

;***************************************************************
; non-specific definitions

VIA_T1CL = VIA+4
VIA_T1LL = VIA+6
VIA_T1LH = VIA+7
VIA_IFR = VIA+13
VIA_IER = VIA+14

ACIA_DATA = ACIA
ACIA_STATUS = ACIA+1
ACIA_COMMAND = ACIA+2
ACIA_CONTROL = ACIA+3

ACIA_RDRF = 8
ACIA_TDRE = 16

;***************************************************************
; standard midi file (SMF) format related routines

init_deltas
	jsr handle_first_events
	dex
	dex
	bne init_deltas

main_loop
	asl clock_pulse
	bcc main_loop
	ror no_active_tracks	; set flag no active tracks
	ldx nbtracks2
track_loop
	lda track_status,x	; is track active ?
	bpl next_track
	lsr no_active_tracks	; reset flag
	sec				; decrement track delta counter
	lda delta_counter,x
	sbc #1
	sta delta_counter,x
	lda delta_counter+1,x
	sbc #0
	sta delta_counter+1,x
	bcs test_counter
	dec delta_high,x
test_counter			; is delta time elapsed ?
	ora delta_counter,x
	ora delta_high,x
	bne next_track

	sec
	ror status_send		; force status_send when changing track
	jsr handle_midi_events
next_track
	dex
	dex
	bne track_loop

	bit checked			; if not already checked,
	bmi test_active
	lda ppqn			; and if less than a quarter note before time,
	cmp delta_counter+6
	lda ppqn+1
	sbc delta_counter+7
	bcc test_active

	ror checked			; then check if next track 3 message is a new line
	ldy #1
	jsr check_text
	bne test_active
	jsr check_line_start
	bne test_active
	jsr clear_line
	stx cursor
	beq next_char
print_string
	lda (track_ptr+6),y
	jsr print_char
next_char
	iny
	dec length
	bne print_string
	dey
skip_delta
	iny
	lda (track_ptr+6),y
	bmi skip_delta
	iny
	jsr check_text
	bne end_display
	jsr check_line_start
	bne print_string
end_display
	lda #0
	sta cursor

test_active
	bit no_active_tracks
	bpl main_loop

wait_completion
	ldy ptr_first
	cpy ptr_last
	bne wait_completion	; wait for pending serial transfers

reset_hardware
	lda #$0A
	sta ACIA_COMMAND	; disable ACIA interrupts
	lda #<10000
	sta VIA_T1LL
	lda #>10000
	sta VIA_T1LH
	sei			; restore old irq vector
	lda irq_save
	sta irq_vector+1
	lda irq_save+1
	sta irq_vector+2
	cli

	jsr clearscreen

clear_vars
	lda #0
	ldx #txtbuflen
clear_vars_loop
	sta txtbuf-1,x
	dex
	bne clear_vars_loop
	rts

;***********************************************************************

handle_midi_events
	jsr handle_one_event
handle_first_events
	lda track_status,x
	bpl no_more_event
	jsr get_delta_time
	beq handle_midi_events	; handle next midi events if 0 delta time
no_more_event
	rts

handle_one_event			; on entry A=event
	jsr inc_ptr
	bpl midi_param
	cmp #$F0
	bcs system_message
						; midi status 8n,9n,An,Bn,Cn,Dn,En
	sta track_status,x
	jsr midi_xmit_status
	lsr status_send
	jsr inc_ptr
midi_param
	lda track_status,x
	bit status_send
	bpl decode_status
	jsr midi_xmit_status
decode_status
	tay
	lda (track_ptr,x)
	cpy #$C0
	bcc send_one_more_param
	cpy #$E0
	bcc send_no_other_param
	bcs send_one_more_param

system_message
	cmp #$FF
	beq meta_event
	jsr midi_xmit
	and #$0F
	beq midi_sysex
	tay
	dey
	beq send_one_param
	dey
	beq send_two_params
	dey
	beq send_one_param
no_param
	rts
midi_sysex
	jsr inc_ptr
	jsr midi_xmit
	cmp #$F7
	bne midi_sysex
	rts
send_two_params
	jsr inc_ptr
send_one_more_param
	jsr midi_xmit
send_one_param
	jsr inc_ptr
send_no_other_param
	jmp midi_xmit

end_of_track
	lsr track_status,x
	rts

meta_event			; pseudo-midi FF command
	jsr inc_ptr
	cpx #2*3		; handle text only in 3rd track
	bne not_text
	cmp #$01		; text message
	beq karaoke_text
not_text
	cmp #$51		; tempo meta-event
	beq tempo_change
	cmp #$2F
	beq end_of_track
					; skip all other FF commands
	jsr inc_ptr
	tay				; get length
	bne skip_loop
	rts

;***********************************************************************
; tempo change

tempo_change
	jsr inc_ptr		; not checked, should be length=3
	tay
tempo_loop
	jsr inc_ptr
	sta tempo-1,y
	dey
	bne tempo_loop
; divide 24-bit tempo by 16-bit ppqn to set the clock tick
setclock
	stx xsave
	lda #0
	sta tmp
	sta tmp+1
	
	ldx #25
	clc
	bcc udiv3
udiv2
	rol tmp
	rol tmp+1
	sec
	lda tmp
	sbc ppqn
	tay
	lda tmp+1
	sbc ppqn+1
	bcc udiv3
	sty tmp
	sta tmp+1
udiv3
	rol tempo
	rol tempo+1
	rol tempo+2
	dex
	bne udiv2
; if the result is greater than 65536, further divide by 16 and
; set a software counter of 16.
	inx
	lda tempo+2
	beq setcount
	ldx #4
shift4
	lsr tempo+2
	ror tempo+1
	ror tempo
	dex
	bne shift4
	ldx #16
setcount
	stx counter_start
	stx counter
	sec			; substract 2 to the count due to timer behavior
	lda tempo
	sbc #2
	sta VIA_T1LL
	lda tempo+1
	sbc #0
	sta VIA_T1LH
	ldx xsave
	rts


;***********************************************************************
; Karaoke text

skip_loop
	jsr inc_ptr
skip_loop2
	dey
	bne skip_loop
	rts

karaoke_text
	jsr inc_ptr
	sta length
	tay
	lsr checked		; reinitializes checked flag for next message
	jsr inc_ptr
	cmp #"@"
	beq special_text

	cmp #"/"		; skip line-start characters
	beq inverse_loop2
	cmp #"\"		; start of lyrics ?
	beq inverse_loop2
	bne inverse_loop1
inverse_loop
	jsr inc_ptr
inverse_loop1
	ldy cursor
	lda (screen_line),y
	ora #$80
	sta (screen_line),y
	tya
	clc
	adc #40
	tay
	lda (screen_line),y
	ora #$80
	sta (screen_line),y
	jsr inc_cursor
inverse_loop2
	dec length
	bne inverse_loop
	rts

special_text
	dey
	jsr inc_ptr
	cmp #"T"
	bne skip_loop2
title_print
	dey
	tya
	eor #$FF
	adc #40
	lsr		; divide by 2 in order to center title
	sta cursor
title
	jsr inc_ptr
	jsr print_char
	dey
	bne title
	jsr newline
	jmp newline

inc_cursor
	ldy cursor
	iny
	cpy #40
	bne *+4
	ldy #80
	sty cursor
	rts

check_text
	lda (track_ptr+6),y
	iny
	cmp #$FF
	bne end_check_text
	lda (track_ptr+6),y
	iny
	cmp #$01
end_check_text
	rts

check_line_start
	lda (track_ptr+6),y
	sta length
	iny
	lda (track_ptr+6),y
	cmp #"/"
	beq lyrics_line
	cmp #"\"
	bne end_check_start
lyrics_line
	lda #<$BF40
	sta screen_line
	lda #>$BF40
	sta screen_line+1
	lda #0			; set Z flag
end_check_start
	rts

clear_line
	ldx #160
	lda #0
clear_loop
	sta $BF40-1,x
	dex
	bne clear_loop
	rts

; translated chars :
; 0 = A
; 1 = E
; 2 = I
; 3 = O
; 4 = U
; 5 = Y
; 6 = N
; 7 = AE
; 8 = O barr
; 9 = 
; A = D barr
; B = no translation
translated_chars
	.db "a","e","i","o","u","y","n",$7B,$60,$7C,"p","."

; accents :
; 0 = rien
; 1 = tilde ()
; 2 = aigu ()
; 3 = grave ()
; 4 = circonflexe ()
; 5 = trma ()
; 6 = rond ()
; 7 = circonflexe invers
; 8 = hampe b,h,k
accents
	.db $20,$7E,$5F,$2B,$5E,$7F,$2F,$2A

; translation table for ISO8859-1, chars 192 to 255 (32 uppercase, 32 lowercase)
translation_table	; high nibble = letter, low nibble = accent
	.db $03,$02,$04,$01,$05,$06,$70,$90
	.db $13,$12,$14,$15,$23,$22,$24,$25
	.db $A0,$61,$33,$32,$34,$31,$35,$B0
	.db $80,$43,$42,$44,$45,$52,$A8,$55

; upper part of lowercase letters
high_part
	.db $20,$20,$23,$20,$24,$20,$25,$20
	.db $23,$2E,$2E,$23,$3D,$20,$20,$20
	.db $20,$20,$20,$20,$3E,$20,$20,$20
	.db $20,$20,$20,$20,$20,$20,$20,$20

accentuated_char
	cmp #$DF	; set carry if lowercase letter
	beq szlig
	php
	and #$1F
	tax
	lda translation_table,x
	pha			; get the character
	lsr
	lsr
	lsr
	lsr
	tax
	lda translated_chars,x
	tay
	pla
	and #$0F	; get the accent
	tax
	lda accents,x
	tax
	tya
	plp
	bcs end_translation
	sbc #$1F	; converts to uppercase
	.db $2C
szlig lda #$5D		;  szlig in special set
	bcs end_translation

; print A at position (screen_line),cursor
; translate ISO8859 to specially defined oric charset
; with bottom part of char in A, upper part in X
print_char
	stx xsave
	sty ysave
	ldx #$20		; default: blank upper part
	cmp #128
	bcs accentuated_char
	cmp #96
	bcc end_translation
	pha
	and #$1F
	tax
	lda high_part,x
	tax
	pla
end_translation
	pha
	txa
	ldy cursor
	sta (screen_line),y
	tya
	clc
	adc #40
	tay
	pla
	sta (screen_line),y
	jsr inc_cursor
end_print
	ldx xsave
	ldy ysave
	rts


;***************************************************************
; helper routines


#if DEBUG

midi_xmit_status
midi_xmit
	rts

#else

midi_xmit_status
	pha
	and #$0F
	sta current_channel
	pla
midi_xmit
	bit melody
	bpl midi_xmit2
	ldy current_channel
	cpy #3		; if melody flag set, don't send on channel 4
	beq end_xmit
midi_xmit2
	ldy ptr_last
	sta buffer,y
	iny
buffer_full
	cpy ptr_first
	beq buffer_full
	sty ptr_last
end_xmit
	rts

#endif

inc_ptr
	inc track_ptr,x
	bne *+4
	inc track_ptr+1,x
	lda (track_ptr,x)
	rts

get_delta_time
	; decodes the delta_time pointed by (track_ptr,x) to delta_counter,x
	; returns Z=1 when delta_time is nul
	; buggy if delta is bigger than 3 bytes (return to bad address)
	ldy #0
	sty delta
push_loop
	iny
	jsr inc_ptr
	asl
	pha
	bcs push_loop
	pla
	sta delta_counter,x
	dey
	beq adjust_delta
	pla
	sta delta
	dey
	beq adjust_delta
	pla
	.db $2C
adjust_delta
	lda #0
	lsr
	lsr
	ror delta
	lsr
	sta delta_high,x
	lda delta
	ror
	sta delta_counter+1,x
	ror delta_counter,x
	ora delta_counter,x
	ora delta_high,x
	rts

;***************************************************************
; interrupt routine

irq_handler
	bit VIA_IFR
	bmi clock_intr
	bit ACIA_STATUS
	bmi acia_intr
end_irq
	rti

clock_intr
	bit VIA_T1CL	; reset T1 interrupt
	dec counter
	bne end_irq
	pha
	lda counter_start
	sta counter
	pla
	sec
	ror clock_pulse
	rti

acia_intr
	pha
	tya
	pha
	lda ACIA_STATUS
	sta status_save
	and #ACIA_RDRF
	beq test_tdr
	lda ACIA_DATA		; midi input not handled
test_tdr
	lda status_save
	and #ACIA_TDRE
	beq end_acia
	ldy ptr_first
	cpy ptr_last
	beq end_acia		; buffer empty
	lda buffer,y
	sta ACIA_DATA
	inc ptr_first
end_acia
	pla
	tay
	pla
	rti

