/* Quick & dirty program to extract midi events in an interval time
 * F.Frances 2004
 * Tempo information is inserted in the file if the interval does not start at 0.
 */

#include <stdio.h>
#include <stdlib.h>

typedef enum { false, true} bool;
unsigned char *track[24];
unsigned char track_header[24][8];
int track_size[24];
int offset_start[24],offset_end[24],offset[24];
int delta[24], start_delay[24];
bool track_finished[24];
int track_status[24];
unsigned char notes_off[4*16];
int tempo=500000;

void usage()
{
	fprintf(stderr,"Usage : midi_extract <midi_file> <start time (seconds)> <end time (seconds)> <extracted_file>\n");
	exit(1);
}

void badformat()
{
	fprintf(stderr,"Bad format\n");
	exit(2);
}

int get_delta(char *ptr)
{
	int delta=0;
	while (*ptr <0) {
		delta|= *ptr&0x7F;
		delta<<=7;
		ptr++;
	}
	delta|= *ptr;
	return delta;
}

void fput_delta(int delta, FILE *fd)
{
	if (delta>=0x200000) fputc((delta>>21)|0x80,fd);
	if (delta>=0x4000) fputc((delta>>14)|0x80,fd);
	if (delta>=0x80) fputc((delta>>7)|0x80,fd);
	fputc(delta&0x7F,fd);
}

int skip_message(int trackno, int offset)
{
	unsigned char byte,length,event;
	while (track[trackno][offset]>=0x80)
		offset++;
	offset++;
	byte=track[trackno][offset];
	if (byte>=0x80)  /* status */
	{
		offset++;
		if (byte<0xF0) {
			track_status[trackno]=byte&0xF0;
		} else {
			switch (byte&0x0F) {
			case 0:	/* system exclusive */
				while(track[trackno][offset++]!=0xF7)
					;
				break;
			case 1:
				offset++;
				break;
			case 2:
				offset+=2;
				break;
			case 3:
				offset++;
				break;
			case 0xF: /* meta event */
				event=track[trackno][offset++];
				if (event==0x2F) {
					track_finished[trackno]=true;
					offset_end[trackno]=offset+1;
				}
				if (event==0x51) {
					tempo= ((track[trackno][offset+1]*256)+track[trackno][offset+2])*256+track[trackno][offset+3];
				}
				length=track[trackno][offset++];
				offset+=length;
				break;
			}
		}
	}
	if (byte<0xF0)
	{
		switch (track_status[trackno]) {
		case 0x80:
		case 0x90:
		case 0xA0:
		case 0xB0:
			offset+=2;
			break;
		case 0xC0:
		case 0xD0:
			offset++;
			break;
		case 0xE0:
			offset+=2;
			break;
		}
	}
	return offset;
}

main(int argc,char *argv[])
{
	unsigned char header[14];
	FILE *in, *out;
	int i,c,nbtracks,newsize,notes_off_sent;
	int started=false,finished=false;
	double pulses=0,start_time=0, end_time=0;
	int ppqn;

	if (argc!=5) usage();
	in=fopen(argv[1],"rb");
	sscanf(argv[2],"%lf",&start_time);
	sscanf(argv[3],"%lf",&end_time);
	out=fopen(argv[4],"wb");
	if (in==NULL || out==NULL || start_time>=end_time) usage();

	fread(header,14,1,in);
	if (strncmp(header,"MThd",4)) badformat();
	nbtracks=header[11];
	ppqn=header[12]*256+header[13];
	for (i=0; i<nbtracks; i++) {
		fread(track_header[i],8,1,in);
		if (strncmp(track_header[i],"MTrk",4)) badformat();
		track_size[i]=track_header[i][7]+256*(track_header[i][6]+256*(track_header[i][5]+256*track_header[i][4]));
		track[i]=(char *)malloc(track_size[i]);
		fread(track[i],track_size[i],1,in);
	}
	fclose(in);


	if (start_time==0) started=true;

	for (i=0;i<nbtracks;i++) {
		delta[i]=get_delta(track[i]+offset[i]);
		if (start_time==0) start_delay[i]=delta[i];
		while (!track_finished[i] && delta[i]==0) {
			offset[i]=skip_message(i,offset[i]);
			delta[i]=get_delta(track[i]+offset[i]);
		}
	}

//	newsize=0;
	while (!finished)
	{
		pulses++;
		if (pulses*tempo>=start_time*ppqn*1E6 && !started) {
			for (i=0;i<nbtracks;i++) {
				offset_start[i]=offset[i];
				start_delay[i]=delta[i]-1;
			}
			started=true;
		}
		if (pulses*tempo>=end_time*ppqn*1E6) break;

//		newsize=sizeof(header)+nbtracks*(sizeof(track_header[0])+4);
		finished=true;
		for (i=0;i<nbtracks;i++) {
			if (!track_finished[i]) {
				finished=false;
				delta[i]--;
				while (!track_finished[i] && delta[i]==0) {
					offset[i]=skip_message(i,offset[i]);
					delta[i]=get_delta(track[i]+offset[i]);
				}
			}
//			newsize+=offset[i];
		}
	}
	notes_off_sent=0;
	fwrite(header,14,1,out);
	for (i=0;i<nbtracks;i++) {
		bool insert_tempo= offset_start[i]!=0;
		if (offset_start[i]==track_size[i])
			offset_start[i]-=4;
		if (!track_finished[i]) {
			track[i][offset[i]++]=0;
			track[i][offset[i]++]=0xFF;
			track[i][offset[i]++]=0x2F;
			track[i][offset[i]++]=0;
		}
		while (track[i][offset_start[i]++]>=0x80)
			;			// skip first delta to replace it
		track_size[i]=offset[i]-offset_start[i];
		if (insert_tempo) track_size[i]+=7;
		if (start_delay[i]<0x80) track_size[i]+=1;
		else if (start_delay[i]<0x4000) track_size[i]+=2;
		else if (start_delay[i]<0x200000) track_size[i]+=3;
		track_header[i][4]=0;
		track_header[i][5]=0;
		track_header[i][6]=track_size[i]>>8;
		track_header[i][7]=track_size[i]&0xFF;
		fwrite(track_header[i],8,1,out);
		if (insert_tempo) {
			fputc(0,out);
			fputc(0xFF,out);
			fputc(0x51,out);
			fputc(0x03,out);
			fputc(tempo>>16,out);
			fputc(tempo>>8,out);
			fputc(tempo,out);
		}
		fput_delta(start_delay[i],out);
		fwrite(track[i]+offset_start[i],offset[i]-offset_start[i],1,out);
	}
	fclose(out);
}
